/** This file is part of Reactor.
 *  Copyright (c) 2020-2022 Kedron Holdings LLC, All Rights Reserved.
 *  Reactor is not public domain or open source. Distribution or derivative works are expressly prohibited.
 */

export const version = 22209;

import api from '/client/ClientAPI.js';

import Reaction from '/client/Reaction.js';

import { idSelector, showSysModal, asyncForEach } from './reactor-ui-common.js';

import { _T } from './i18n.js';

import ReactionEditor from "./reaction-editor.js";

export default (function($) {

    var tabId = "tab-reactions";
    var editor = false;

    var toggleDetail = async function( event ) {
        let $row = $( event.currentTarget ).closest( 'div.re-reaction' );
        var id = $row.attr( 'id' );
        var $el = $( '#detail-' + id + '.collapse' );
        if ( ! $el.hasClass( 'loaded' ) ) {
            var reaction = await Reaction.getInstance( id );
            if ( reaction ) {
                var $body = $( '.card-body', $el ).empty();
                if ( 0 === (reaction.actions || []).length ) {
                    $body.text( _T('This reaction is empty.') );
                } else {
                    var $list = $( '<ol></ol>' ).appendTo( $body );
                    await asyncForEach( reaction.actions || [], async ( action ) => {
                        try {
                            let str = await ReactionEditor.makeActionDescription( action );
                            $( '<li></li>' ).text( str ).appendTo( $list );
                        } catch( e ) {
                            console.log( e );
                            $( '<li></li>' ).text( "ERROR" ).appendTo( $list );
                        }
                    });
                }
                $( '<div></div>' ).text( _T('Reaction ID: {0}', reaction.id) ).appendTo( $body );
                $el.addClass( 'loaded' );
            } else {
                return;
            }
        }
        if ( $el.hasClass( 'show' ) ) {
            $el.slideUp( 250 ).removeClass( 'show' );
            $( '.re-reaction-name > i', $row ).removeClass( 'bi-chevron-down' ).addClass( 'bi-chevron-right' );
        } else {
            $el.slideDown( 250 ).addClass( 'show' );
            $( '.re-reaction-name > i', $row ).removeClass( 'bi-chevron-right' ).addClass( 'bi-chevron-down' );
        }
        return false;
    };

    var editReaction = async function( reaction, isNew ) {
        var $tab = $( 'div#' + tabId );
        $tab.empty();

        let newReaction = !!isNew;

        var $ct = $('<div class="container-fluid re-editor-controls"></div>' ).appendTo( $tab );
        var $row = $( '<div class="row"></div>' ).appendTo( $ct );
        $( '<div class="col"><h1></h1></div>' ).appendTo( $row );
        $( `<div class="col text-end mt-3">
  <button class="btn btn-sm btn-success saveconf savecont">${_T(['#edit-button-save','Save'])}</button>
  <button class="btn btn-sm btn-danger revertconf saveexit">${_T(['#edit-button-exit','Exit'])}</button>
</div>` ).appendTo( $row );
        $( 'btn', $row ).prop( 'disabled', true );
        $( 'h1', $row ).text( reaction.name || reaction.id ).on( 'click', function( ev ) {
            var $el = $( ev.currentTarget );
            var $form = $( '<div class="form-group"><input id="newname" class="form-control form-control-sm"></div>' );
            $( 'input', $form ).val( $el.text() );
            showSysModal({
                title: _T(['#reaction-rename-title','Rename Reaction']),
                body: $form,
                buttons: [{
                    label: _T(['#reaction-rename-button-cancel','Cancel']),
                    class: "btn-primary",
                    close: true
                },{
                    label: _T(['#reaction-rename-button-rename','Rename']),
                    class: "btn-success",
                    event: "rename"
                }
            ]
            }).then( data => {
                if ( "rename" === data ) {
                    var newname = $( '#sysdialog #newname' ).val() || "";
                    if ( "" !== newname && newname !== reaction.name ) {
                        editor.rename( newname );
                        $el.text( newname );
                    }
                }
            });
        });

        $tab.data( 'mode', 'edit' ).attr( 'data-mode', 'edit' );
        editor = new ReactionEditor( reaction, $tab, { fixedName: reaction.name || reaction.id } );
        editor.editor().on( 'update-controls', function( ev, ui ) {
            /* During editing, various changes make the reaction saveable, or not. */
            console.log("got update-controls");
            $( 'button.saveconf', $tab ).prop( 'disabled', ui.errors || !ui.modified );
            $( 'button.revertconf', $tab ).prop( 'disabled', false );
        }).on( 'close', function() {
            console.log("got close");
            $tab.empty();
            $tab.data( 'mode', '' ).attr( 'data-mode', '' );
            $tab.trigger( 'activate' );
        }).on( 'modified', function( ev, ui ) {
            /* Sent when the cached reaction is modified (no errors) */
            console.log("got modified");
            $( 'button.saveconf', $tab ).prop( 'disabled', ui.errors || !ui.modified );
            $( 'button.revertconf', $tab ).prop( 'disabled', false );
        }).on( 'ready', function() {
            $( 'button.saveconf', $ct ).on( 'click', async function( event ) {
                var $el = $( event.currentTarget );
                if ( !editor.canSave() ) {
                    showSysModal( {
                        title: 'Uncorrected errors remain!',
                        body: 'Please correct the indicated errors before saving.'
                    } );
                    return;
                }
                $( 'button.saveconf', $tab ).prop( 'disabled', true );
                var reaction = editor.data;
                var robj = await Reaction.getInstance( reaction.id );
                var dobj = robj.getDataObject();
                reaction.serial = ( reaction.serial || 0 ) + 1;
                reaction.mdt = Date.now();
                dobj.setValue( reaction );
                dobj.save().then( function() {
                    editor.notifySaved();
                    newReaction = false;
                    if ( $el.hasClass( 'saveexit' ) ) {
                        editor.close();
                    }
                }).catch( function( err ) {
                    console.log("save failed",err);
                    showSysModal( {
                        title: _T('#dialog-error-title'),
                        body: _T('The save failed; try again. {0}')
                    } );
                    editor.updateSaveControls();
                });
            });
            $( 'button.revertconf', $ct ).on( 'click', async function( event ) {
                var reaction = editor.data;
                var robj = await Reaction.getInstance( reaction.id );
                if ( editor.isModified() ) {
                    showSysModal( {
                        title: _T(['#reactionedit-unsaved-title','Reaction Modified (Unsaved Changes)']),
                        body: _T(['#reactionedit-unsaved-prompt','You have unsaved changes to this reaction. Really discard changes and exit?']),
                        buttons: [{
                            label: _T(['#reactionedit-unsaved-button-discard','Discard Changes']),
                            event: "discard",
                            class: "btn-danger"
                        },{
                            label: _T(['#reactionedit-unsaved-button-cancel','Cancel']),
                            close: true,
                            class: "btn-success"
                        }]
                    }).then( async data => {
                        if ( "discard" === data ) {
                            if ( newReaction ) {
                                await robj.delete();
                            }
                            editor.close();
                        }
                    });
                } else {
                    if ( newReaction ) {
                        await robj.delete();
                    }
                    editor.close();
                }
            });
        });
        editor.edit();
        if ( newReaction ) {
            editor.signalModified();
        }
    };

    var setEnabledToggle = function( $btn, state ) {
        if ( state ) {
            $( 'i', $btn ).removeClass( 'bi-toggle-off' ).addClass( 'bi-toggle-on' );
            $btn.removeClass( 'text-danger' ).addClass( 'text-success' )
                .attr( 'title', _T("Enabled; click to disable") );
        } else {
            $( 'i', $btn ).removeClass( 'bi-toggle-on' ).addClass( 'bi-toggle-off' );
            $btn.removeClass( 'text-success' ).addClass( 'text-danger' )
                .attr( 'title', _T("Disabled; click to enable") );
        }
    };

    var getReactionRow = function( reaction ) {
        var $row = $( '<div class="row re-listport-item re-reaction"></div>' ).attr( 'id', reaction.id );
        var $c = $( '<div class="col"></div>' )
            .appendTo( $row );
        $( `<button class="btn bi-btn me-2 text-success" title="${_T(['#reaction-action-run','Run Now'])}" data-action="run"><i class="bi bi-play"></i></button>` )
            .appendTo( $c );
        var $e = $( '<span class="re-reaction-name"></span>' )
            .text( reaction.name || reaction.id )
            .attr( 'title', String( reaction.name ) + " ID " + reaction.id )
            .appendTo( $c );
        $( '<i class="bi bi-chevron-right ms-1"></i>' ).appendTo( $e );

        $c = $( '<div class="col text-end re-reaction-controls"></div>' )
            .appendTo( $row );
        $( '<button class="btn bi-btn ms-2 text-success" data-action="disable" title="Enable/Disable Reaction"><i class="bi bi-toggle-on"></i></button>' )
            .appendTo( $c );
        $( `<button class="btn bi-btn ms-2 text-primary" data-action="edit" title="${_T(['#reaction-action-edit','Edit this Reaction'])}"><i class="bi bi-pencil"></i></button>` )
            .appendTo( $c );
        $( `<button class="btn bi-btn ms-2 text-secondary" data-action="copy" title="${_T(['#reaction-action-copy','Copy Reaction'])}"><i class="bi bi-share-fill"></i></button>` )
            .appendTo( $c );
        $( `<button class="btn bi-btn ms-4 text-danger" data-action="delete" title="${_T(['#reaction-action-delete','Delete this Reaction'])}"><i class="bi bi-x"></i></button>` )
            .appendTo( $c );

        /* Detail card (collapsible) */
        $( '<div class="w-100"></div>').appendTo( $row );
        $( '<div class="col-xs-12 col-sm-12 collapse"><div class="card card-body">Jeffrey Epstein didn\'t kill himself.</div></div>' )
            .attr( 'id', 'detail-' + reaction.id )
            .appendTo( $row );

        var $el = $( 'button[data-action="disable"]', $row );
        setEnabledToggle( $el, !reaction.disabled );

        $( 'span.re-reaction-name', $row ).on( 'click', toggleDetail );

        $( 'button[data-action="run"]', $row ).on( 'click', function( event ) {
            var $el = $( event.currentTarget );
            var $row = $el.closest( '.re-reaction' );
            var rid = $row.attr( 'id' );
            $row.stop().clearQueue().css( 'background-color', '#0c0' );
            api.startReaction( rid ).catch( function() {} ).finally( function() {
                $row.animate( { backgroundColor: '#fff' }, 2000);
            });
        });

        $( 'button[data-action="disable"]', $row ).on( 'click', async function( event ) {
            var $el = $( event.currentTarget );
            var rid = $el.closest( '.re-reaction' ).attr( 'id' );
            var robj = await Reaction.getInstance( rid );
            var dobj = robj.getDataObject();
            dobj.value.disabled = ! dobj.value.disabled;
            if ( false === dobj.value.disabled ) {
                delete dobj.value.disabled;
            }
            dobj.save();
            setEnabledToggle( $el, !robj.disabled );
            return false;
        });

        $( 'button[data-action="copy"]', $row ).on( 'click', function( event ) {
            var $el = $( event.currentTarget );
            var $row = $el.closest( '.re-reaction' );
            var rid = $row.attr( 'id' );
            Reaction.getInstance( rid ).then( src => {
                Reaction.create().then( dest => {
                    var dobj = dest.getDataObject();
                    dobj.value.name = src.name + ' Copy';
                    dobj.value.serial = 1;
                    dobj.value.actions = [ ...src.actions ];
                    dobj.save().then( () => {
                        getReactionRow( dest ).insertAfter( $row );
                    });
                });
            });
            return false;
        });

        $( 'button[data-action="delete"]', $row ).on( 'click', async function( event ) {
            var $el = $( event.currentTarget );
            var $row = $el.closest( '.re-reaction' );
            var rid = $row.attr( 'id' );
            var robj = await Reaction.getInstance( rid );
            showSysModal( {
                title: _T(['#reaction-delete-title','Confirm Deletion']),
                body: _T(['#reaction-delete-prompt','Really delete {0:q}? There is no "undo"'], robj.name),
                buttons: [
                    {
                        label: _T(['#reaction-delete-button-delete','Delete']),
                        class: "btn-danger",
                        event: "delete"
                    },
                    {
                        label: _T(['#reaction-delete-button-cancel','Cancel']),
                        class: "btn-success"
                    }
                ]
            } ).then( async function( data ) {
                if ( "delete" === data ) {
                    robj.delete().then( () => {
                        /* Remove the displayed row */
                        $row.remove();
                    });
                }
                return data;
            } );
            return false;
        } );

        $( 'button[data-action="edit"]', $row ).on( 'click', async function( ev ) {
            var rid = $( ev.target ).closest( 'div.re-reaction' ).attr( 'id' );
            var reaction = await Reaction.getInstance( rid );
            if ( ! reaction ) {
                console.log("Reaction " + String(rid) + " not found");
                return;
            }
            editReaction( reaction.getDataObject().value );
        });

        return $row;
    };

    var activate = async function( event ) {
        console.log("ReactorUIReactions.show() running");
        console.log(event);
        var $tab = $( event.target );
        if ( tabId !== $tab.attr( 'id' ) || !$tab.hasClass( "re-tab-container" ) ) {
            console.log( "    event is not for the tab, but a child" );
            return;
        }
        /* If editing, don't paint this, it's just an event that bubbled up to us */
        if ( "edit" === $tab.data( 'mode' ) ) {
            console.log("Stopping event in edit mode");
            return false;
        }
        $tab.empty();
        $tab.data( 'mode', 'list' ).attr( 'data-mode', 'list' );
        var $ct = $( '<div class="re-listport container-fluid ps-0"></div>' ).appendTo( $tab );

        $( `<div class="re-listport-head container-fluid ps-0">
  <div class="row align-middle">
    <div class="col align-top">
      <span class="re-listport-title">${_T(['#reactions-page-title','Global Reactions'])}</span>
    </div>
    <div class="col-auto float-end mt-2">
      <button id="newreaction" class="btn btn-sm btn-success">${_T(['#reactions-button-create','Create Reaction'])}</button>
    </div>
  </div>
</div>` )
            .appendTo( $ct );

        var $body = $( '<div class="re-listport-body container-fluid ps-0"></div>' )
            .appendTo( $ct );

        var reactions = await Reaction.getGlobalReactions();
        reactions.sort( function( a, b ) {
            return (a.name || a.id).localeCompare( b.name || b.id, undefined, { sensitivity: 'base' } );
        });
        var nrs = reactions.length;
        /* ??? better to use ul/li ? */
        for ( var k=0; k<nrs; ++k ) {
            var reaction = reactions[k];
            var $row = getReactionRow( reaction );
            $row.appendTo( $body );
        }

        $( 'button#newreaction' ).on( 'click', async function() {
            var reaction = await Reaction.create();
            editReaction( reaction.getDataObject().value, true );
        });

        return false;
    };

    var suspend = function( event ) {
        console.log("ReactorUIReactions.show() running");
        console.log(event);
        var $tab = $( event.target );
        if ( tabId !== $tab.attr( 'id' ) || !$tab.hasClass( "re-tab-container" ) ) {
            console.log( "    event is not for the tab, but a child" );
            return;
        }
        /* If editing, don't paint this, it's just an event that bubbled up to us */
        if ( "edit" === $tab.data( 'mode' ) ) {
            if ( editor && editor.isModified() ) {
                console.log( "Stopping event in edit mode" );
                return false;
            }
        }
        $tab.data( 'mode', '' ).attr( 'data-mode', '' ).empty();
    };

    return {
        "init": function( $main ) {
            $( 'div#' + idSelector( tabId ) + '.re-tab-container' ).remove();
            var $tab = $( '<div></div>' )
                .hide()
                .attr( 'id', tabId )
                .addClass( 're-tab-container' )
                .appendTo( $main );
            $tab.on( 'activate', activate );
            $tab.on( 'suspend', suspend );

            /* Our styles. */
            if ( 0 === $('link#reactor-reaction-styles').length ) {
                $('head').append('<link id="reactor-reaction-styles" href="lib/css/reactor-ui-reactions.css" rel="stylesheet">');
            }
        }
    };
})(jQuery);
