/** This file is part of Reactor.
 *  Copyright (c) 2020-2022 Kedron Holdings LLC, All Rights Reserved.
 *  Reactor is not public domain or open source. Distribution or derivative works are expressly prohibited.
 */

export const version = 22206;

import api from '/client/ClientAPI.js';
import Rulesets from '/client/Rulesets.js';
import Rule from '/client/Rule.js';

import { isEmpty } from './reactor-ui-common.js';

import { _T } from './i18n.js';

/* ??? needs conversion to Tab class */

export default (function($) {

    var scopeTimer = false;
    var scopeInterval = 200;
    var scopeScale = 4;
    var traceHeight = 64;
    var traceMargin = 8;
    var tracePadding = 4;
    var scopeTraces = [];

    var isRunning = function() {
        return scopeTimer ? true : false;
    };

    /** Round to a pleasant-looking value */
    var prettyFloat = function( n ) {
        let str = Number( n ).toPrecision( 4 );
        if ( str.indexOf( '.' ) >= 0 ) {
            str = str.replace( /0+$/, "" ).replace( /\.$/, "" );
        }
        return str;
    };

    var resizeCanvas = function() {
        var $tab = $( 'div#tab-scope.re-tab-container' );
        var $canvas = $( 'canvas#scope', $tab );
        var $div = $canvas.closest( 'div' );
        var width = Math.floor( $div.innerWidth() );
        var height = scopeTraces.length * ( traceHeight + 2*tracePadding ) + /* aka boxHeight */
            (scopeTraces.length-1) * traceMargin;
        height = Math.max( 240, height );
        $canvas.css( 'width', width+'px' ).css( 'height', height+'px' );
        $canvas.get(0).width = width;
        $canvas.get(0).height = height;
        return [ width, height ];
    };

    var nice = function( minmax ) {
        const steps = [ 1, 2, 3, 4, 5, 10 ];
        let range = minmax[1] - minmax[0];
        if ( range < 0.01 ) {
            range = 1.0;
        }
        let tickstep = range / 5.0; /* five "tick" (bottom, bottom center, midline, top center, top) */
        let mag = Math.floor( Math.log10( tickstep ) );
        let radj = tickstep / ( 10 ** mag );
        let step = steps.find( v => ( v >= radj ? v : false ) );
        step =  step * 10 ** mag;
        let nlo = Math.floor( minmax[0] / step ) * step;
        let nhi = Math.ceil( minmax[1] / step ) * step;
        if ( nhi <= nlo ) {
            nhi = nlo + step;
        }
        return [nlo,nhi];
    };

    var updateScope = async function() {
        /* Each pixel on the canvas is a measurement period, with 0 being the newest */
        var t0 = Date.now();
        var $canvas = $( 'canvas#scope' );
        var canvasElement = $canvas.get( 0 );
        var ctx = canvasElement.getContext( "2d" );
        var m = resizeCanvas();
        var width = m[0];
        // var height = m[1];
        var nsample = Math.floor( width / scopeScale );
        for ( var k=0; k<scopeTraces.length; k++ ) {
            var tr = scopeTraces[ k ];
            var val, caption;
            if ( tr.entity ) {
                /* Entity trace */
                var e = api.getEntity( tr.entity );
                if ( e ) {
                    caption = e.getName() + "." + tr.attr;
                    val = e.getAttribute( tr.attr );
                    if ( "boolean" === typeof val ) {
                        val = val ? 1 : 0;
                        tr.scale = [0,1];
                        tr.binary = true;
                    } else if ( null !== val ) {
                        val = parseFloat( val );
                        if ( isNaN( val ) ) {
                            val = null;
                        }
                    }
                } else {
                    val = null;
                }
            } else if ( tr.rule ) {
                /* Rule trace */
                try {
                    var rule = await Rule.getInstance( tr.rule );
                    var states = await rule.getStates();
                    if ( states ) {
                        val = (states.rule || {}).evalstate;
                        if ( null !== val ) {
                            val = val ? 1 : 0;
                        }
                        tr.scale = [0,1];
                        tr.binary = true;
                    } else {
                        val = null;
                    }
                    caption = rule.name;
                } catch( err ) {
                    console.log("Scope: can't retrieve state for rule", tr.rule, err);
                    val = null;
                }
            } else {
                /* Hmmmm.... */
                val = null;
                caption = "?";
            }

            /* Insert value */
            tr.values.unshift( val );
            while ( tr.values.length > nsample ) {
                tr.values.pop();
            }

            /* Clean up scale */
            if ( null !== val && !tr.binary ) {
                let lo = Math.min( ...tr.values );
                let hi = Math.max( ...tr.values );
                let s = nice( [lo, hi] );
                if ( tr.scale[0] === null || tr.scale[1] === null ) {
                    tr.scale = s;
                } else if ( ! ( s[0] >= tr.scale[0] && s[1] <= tr.scale[1] ) ) {
                    tr.scale = s;
                }
            }

            /* Redraw the trace */
            var boxHeight = traceHeight + 2*tracePadding;
            var yorg = k * (boxHeight + traceMargin);
            var ybot = yorg + boxHeight;
            ctx.clearRect( 0, yorg, width, boxHeight );
            ctx.fillStyle = '#ffffff';
            ctx.fillRect( 0, yorg, width, boxHeight );
            ctx.strokeStyle = '#999';
            ctx.beginPath();
            ctx.moveTo( 0, ybot );
            ctx.lineTo( width, ybot );
            ctx.lineTo( width, ybot - boxHeight );
            ctx.lineTo( 0, ybot - boxHeight );
            // ctx.lineTo( 0, ybot );
            ctx.closePath();
            ctx.stroke();
            ctx.beginPath();
            ctx.moveTo( 0, ybot - boxHeight/2 );
            ctx.lineTo( width, ybot - boxHeight/2 );
            // ctx.closePath();
            ctx.stroke();
            ctx.strokeStyle = '#090';
            ctx.fillStyle = '#000';
            ctx.font = "9pt Arial";
            ctx.fillText( caption, 72, ybot - 8);
            var mn = tr.scale[0] === null ? 0 : tr.scale[0];
            var mx = tr.scale[1] === null ? 1 : tr.scale[1];
            ctx.fillText( prettyFloat( mn ), 8, ybot - 8 );
            ctx.fillText( prettyFloat( mx ), 8, ybot - boxHeight + 16 );
            var scale = traceHeight / ( mx - mn );
            var nextMove = true;
            for ( var j=0; j<tr.values.length; ++j ) {
                if ( tr.values[j] === null ) {
                    /* Finish existing stroke, if any */
                    if ( !nextMove ) {
                        ctx.stroke();
                    }
                    nextMove = true;
                } else {
                    var y1 = ybot - Math.floor( (tr.values[j] - mn) * scale + 0.5 ) - tracePadding;
                    if ( nextMove ) {
                        /* Start new stroke */
                        ctx.beginPath();
                        ctx.moveTo( width-j*scopeScale, y1 );
                        nextMove = false;
                    } else {
                        ctx.lineTo( width-j*scopeScale, y1 );
                    }
                }
            }
            if ( !nextMove ) {
                /* Finish unfinished stroke */
                ctx.stroke();
            }
        }

        if ( scopeTraces.length > 0 ) {
            var dt = Date.now() - t0;
            scopeTimer = setTimeout( updateScope, scopeInterval - Math.min( dt, scopeInterval ) );
        } else {
            scopeTimer = false;
        }
    };

    function handleAddTraceClick( event ) {
        // var $el = $( event.currentTarget );
        var $fs = $( 'select#source-obj' );
        var objid = $fs.val() || "";
        if ( isEmpty( objid ) ) {
            return;
        }
        var $opt = $( 'option[value="' + objid + '"]' );
        if ( "entity" === $opt.data( 'source' ) ) {
            $fs = $( 'select#source-sub' );
            var subid = $fs.val() || "";
            if ( isEmpty( subid ) ) {
                return;
            }
            scopeTraces.push( { x:0, entity: objid, attr: subid, values:[], scale: [null,null] } );
        } else if ( "rule" === $opt.data( 'source' ) ) {
            scopeTraces.push( { x:0, rule: objid, values: [], scale: [null,null] } );
        }
        if ( ! isRunning() ) {
            scopeTimer = setTimeout( updateScope, scopeInterval );
        }
    }

    function handleSourceSubChange( event ) {
        var $el = $( event.currentTarget );
        var subid = $el.val() || "";
        $( 'button#addtrace' ).prop( 'disabled', isEmpty( subid ) );
    }

    function handleSourceChange( event ) {
        var $el = $( event.currentTarget );
        var objid = $el.val() || "";
        if ( isEmpty( objid ) ) {
            return;
        }
        var $opt = $( 'option[value="' + objid + '"]' );
        var $fs = $( 'select#source-sub' ).empty();
        if ( "entity" === $opt.data( 'source' ) ) {
            $fs.prop( 'disabled', false );
            $( 'button#addtrace' ).prop( 'disabled', false );
            var e = api.getEntity( objid );
            $( '<option></option>' ).val( "" ).text( _T('(choose attribute)') )
                .appendTo( $fs );
            for ( var attr in e.getAttributes() || {} ) {
                $( '<option></option>' ).val( attr ).text( attr )
                    .appendTo( $fs );
            }
            $fs.val( e.getPrimaryAttribute() || "" );
        } else {
            /* Rule */
            $fs.prop( 'disabled', true );
            $( 'button#addtrace' ).prop( 'disabled', false );
        }
    }

    var activate = async function( event ) {
        // console.log("ReactorUIScope.show() running");
        // console.log(event);
        var $tab = $( event.target );
        if ( "tab-scope" !== $tab.attr( 'id' ) || !$tab.hasClass( "re-tab-container" ) ) {
            // console.log( "    event is not for the tab, but a child" );
            return;
        }

        /* If the scope is not running, redraw tab. */
        if ( ! isRunning() ) {
            /* Redraw tab */
            $tab.empty();
            var $row = $( '<div class="row"></div>' ).appendTo( $tab );
            var $col = $( '<div class="col"></div>' ).appendTo( $row );
            var $div = $( '<div class="form-group form-inline" ></div>' ).appendTo( $col );
            var $el = $( '<select id="source-obj" class="form-select form-select-sm"></select>' );
            $( '<option></option>' ).val( "" ).text( _T('(choose target)') )
                .appendTo( $el );
            var $og = $( '<optgroup></optgroup>' ).attr( 'label', _T('Entities') )
                .appendTo( $el );
            var elist = api.getEntities().sort( function( a, b ) {
                return a.getName().localeCompare( b.getName(), undefined, { sensitivity: 'base' } );
            });
            for ( var k=0; k<elist.length; ++k ) {
                var e = elist[k];
                if ( e.getType().match( /^(System|Group|Script)/ ) ) {
                    continue;
                }
                $( '<option></option>' ).val( e.getCanonicalID() )
                    .text( e.getName() + ' (' + e.getCanonicalID() + ')' )
                    .data( 'source', 'entity' ).attr( 'data-source', 'entity' )
                    .appendTo( $og );
            }

            let first = true;
            let rulesets = await Rulesets.getRulesets();
            rulesets.forEach( set => {
                if ( set.rules.length > 0 ) {
                    if ( first ) {
                        $( '<optgroup></optgroup>' ).attr( 'label', '' )
                            .addClass( 'divider' )
                            .appendTo( $el );
                        first = false;
                    }
                    var $og = $( '<optgroup></optgroup>' ).attr( 'label', set.name )
                        .appendTo( $el );
                    set.getRules().then( rules => {
                        rules.forEach( rule => {
                            $( '<option></option>' ).val( rule.id ).text( rule.name )
                                .data( 'source', 'rule' ).attr( 'data-source', 'rule' )
                                .appendTo( $og );
                        });
                    });
                }
            });

            $el.on( 'change', handleSourceChange ).appendTo( $div );
            $el = $( '<select id="source-sub" class="form-select form-select-sm"></select>' );
            $el.on( 'change', handleSourceSubChange ).appendTo( $div );
            $( `<button id="addtrace" class="btn btn-sm btn-outline-success">${_T('Add Trace')}</button>` )
                .on( 'click', handleAddTraceClick ).appendTo( $div );
            $( `<button id="cleartraces" class="btn btn-sm btn-outline-danger">${_T('Clear Traces')}</button>` )
                .appendTo( $div )
                .on( 'click', function() {
                    scopeTraces = [];
                    if ( scopeTimer ) {
                        clearTimeout( scopeTimer );
                        scopeTimer = false;
                    }
                    resizeCanvas();
                });

            $el = $( '<div class="w-100"></div>' ).appendTo( $tab );
            $( '<canvas id="scope" width="200" height="200"></canvas>' )
                .appendTo( $el );
            setTimeout( resizeCanvas, 100 );

            $( 'select#source' ).val( "" ).trigger( 'change' );
        }
    };

    var suspend = function( event ) {
        // console.log("ReactorUIScope.hide() running");
/*
        console.log(event);
        var $tab = $( event.target );
        if ( "tab-scope" !== $tab.attr( 'id' ) || !$tab.hasClass( "re-tab-container" ) ) {
            console.log( "    event is not for the tab, but a child" );
            return;
        }
*/

        /**
         * For this function, we do not clear the tab, remove data, or
         * stop running updates. We want the scope to continue running
         * in the background.
         */
    };

    return {
        "init": function( $main ) {
            $( 'div#tab-scope.re-tab-container' ).remove();
            var $tab = $( '<div></div>' )
                .hide()
                .attr( 'id', 'tab-scope' )
                .addClass( 're-tab-container' )
                .appendTo( $main );
            $tab.on( 'activate', activate );
            $tab.on( 'suspend', suspend );

            /* Our styles. */
            if ( 0 === $('link#reactor-scope-styles').length ) {
                $('head').append('<link id="reactor-scope-styles" href="lib/css/reactor-ui-scope.css" rel="stylesheet">');
            }
        }
    };
})(jQuery);
